unit Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, StdCtrls, Grids, ExtCtrls, ComCtrls,
  HashAlgUnified_U, XPMan, ActnList;

type
  TfrmMain = class(TForm)
    sgFilesList: TStringGrid;
    MainMenu: TMainMenu;
    miFile: TMenuItem;
    miNew: TMenuItem;
    miOpen: TMenuItem;
    miSave: TMenuItem;
    miSaveAs: TMenuItem;
    N1: TMenuItem;
    miExit: TMenuItem;
    OpenDialog: TOpenDialog;
    OpenDialogAdd: TOpenDialog;
    SaveDialog: TSaveDialog;
    gbStatus: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    PopupMenu1: TPopupMenu;
    edStatusOK: TEdit;
    edStatusChanged: TEdit;
    edStatusUnchecked: TEdit;
    edStatusUnknown: TEdit;
    miCheckSelected2: TMenuItem;
    miCheckAll2: TMenuItem;
    Check1: TMenuItem;
    miCheckSelected1: TMenuItem;
    miCheckAll1: TMenuItem;
    N2: TMenuItem;
    miRecalcSelected: TMenuItem;
    miRecalcAll: TMenuItem;
    miHelp: TMenuItem;
    miAbout: TMenuItem;
    miAddFile1: TMenuItem;
    miAddDir1: TMenuItem;
    N3: TMenuItem;
    N4: TMenuItem;
    miProperties2: TMenuItem;
    Label5: TLabel;
    edHashType: TEdit;
    N5: TMenuItem;
    miRemoveSelected2: TMenuItem;
    N6: TMenuItem;
    miRemoveSelected1: TMenuItem;
    miProperties1: TMenuItem;
    edStatusDeleted: TEdit;
    Label6: TLabel;
    pnlLine1: TPanel;
    Label7: TLabel;
    edStatusTotal: TEdit;
    Label8: TLabel;
    edBytesTotal: TEdit;
    statusBar: TStatusBar;
    ProgressBar: TProgressBar;
    N8: TMenuItem;
    miAddFile2: TMenuItem;
    miAddDir2: TMenuItem;
    miView: TMenuItem;
    miReport: TMenuItem;
    pnlLine2: TPanel;
    pnlLine3: TPanel;
    N7: TMenuItem;
    miOptions: TMenuItem;
    XPManifest1: TXPManifest;
    ActionList1: TActionList;
    actAddFile: TAction;
    actAddDir: TAction;
    actRemoveSelected: TAction;
    actCheckSelected: TAction;
    actCheckAll: TAction;
    actProperties: TAction;
    procedure miExitClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure miSaveClick(Sender: TObject);
    procedure miSaveAsClick(Sender: TObject);
    procedure miOpenClick(Sender: TObject);
    procedure sgFilesListDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure RecalcSelectedClick(Sender: TObject);
    procedure RecalcAllClick(Sender: TObject);
    procedure miAboutClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure sgFilesListClick(Sender: TObject);
    procedure statusBarDrawPanel(StatusBar: TStatusBar;
      Panel: TStatusPanel; const Rect: TRect);
    procedure sgFilesListDblClick(Sender: TObject);
    procedure miNewClick(Sender: TObject);
    procedure miReportClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure miOptionsClick(Sender: TObject);
    procedure FormClick(Sender: TObject);
    procedure actAddFileExecute(Sender: TObject);
    procedure actAddDirExecute(Sender: TObject);
    procedure actRemoveSelectedExecute(Sender: TObject);
    procedure actCheckSelectedExecute(Sender: TObject);
    procedure actCheckAllExecute(Sender: TObject);
    procedure actPropertiesExecute(Sender: TObject);
  private
    Silent: boolean;
    VerySilent: boolean;

    CancelFlag: boolean;

    Hasher: THashAlgUnified;
    Key: string;

    fReportTitle: string;
    fReportBody: TStringList;
    fReportLastOpStartTime: TDateTime;
    fReportLastOpEndTime: TDateTime;

    fFileChanged: boolean;

    Filenames: TStringList;
    FileSizes: TStringList;
    FileHashes: TStringList;
    FileStatus: TStringList;

    fLoaded: boolean;
    fListFilename: string;
    fListHashType: string;

    function  GetSize(filename: string): int64;
    function  GetChecksumFile(filename: string): string;
    function  GetChecksumString(data: string): string; overload;
    function  GetChecksumString(data: string; checksumKey: string): string; overload;

    function  AddFile(filename: string): boolean;
    procedure AddToDisplay(filename, size, hash, status: string);

    procedure ClearList();
    function  StripLine(var lines: TStringList): string;
    function  GetValue(name: string; nameValue: string): string;
    function  SaveLoadList(filename: string; doSave: boolean; hmacKey: string = ''): boolean;

    function  Save(): boolean;
    function  SaveAs(): boolean;
    procedure Open();

    procedure SetFileChanged(changed: boolean);
    function  GetFileChanged(): boolean;

    function  GetListFilename(): string;
    procedure SetListFilename(filename: string);
    function  GetListHashType(): string;
    function  SetListHashType(hashType: string): boolean;

    procedure DrawTheText(AGrid: TStringGrid; ARect: TRect; AValue: string; Alignment: TAlignment);

    procedure UpdateDisplayStatus();

    procedure EnableDisable();

    procedure DeleteRow(row: integer);

    procedure CheckRange(startIdx, endIdx: integer);
    procedure RecalcRange(startIdx, endIdx: integer);

    procedure ReportNew(title: string);
    procedure ReportAddFileList(files: TStringList; title: string; msgSeg: string);
    procedure ReportAddLine(line: string);
    procedure ReportAddTimeTaken();
    procedure ReportShow();

    procedure StartOperation(operation: string);
    function  RunningOperation(): boolean;
    procedure EndOperation();

    procedure ProgressBarReset(max: int64);
    procedure ProgressBarTick();

    function  UserConfirmCancelOp(): boolean;

    function  GenerateRandomString(minBits: integer): string;
    function  GetKeyFromUser(var userKey: string): boolean;

  public
    { Public declarations }
  end;

var
  frmMain: TfrmMain;

implementation

{$R *.DFM}

// The various icons (e.g. for file association)
{$R Icons.DCR}

uses
  Math,
  inifiles,
  SDUGeneral,
  SDUFileIterator_U,
  HMAC,
  HashValue_U,
  HashAlg_U,
  About_U,
  AddDir,
  Options,
  Properties,
  Report,
  SelectHash,
  KeyEntry;

const
  CRLF = #10+#13;

  statusOK        = 'OK';
  statusChanged   = 'Changed';
  statusDeleted   = 'Deleted';
  statusUnchecked = 'Unchecked';
  statusUnknown   = 'Unknown';

  operationReady = 'Ready';

  colFilename = 0;
  colSize     = 1;
  colStatus   = 2;

procedure TfrmMain.miExitClick(Sender: TObject);
begin
  Close();

end;


procedure TfrmMain.FormCreate(Sender: TObject);
var
  loadFile: string;
  hmacKey: string;
begin
  FLoaded := FALSE;

  self.Caption := Application.Title;

  Hasher:= THashAlgUnified.Create(nil);

  Filenames := TStringList.Create();
  FileSizes := TStringList.Create();
  FileHashes := TStringList.Create();
  FileStatus := TStringList.Create();

  fReportBody:= TStringList.Create();
  fReportTitle := '';

  edHashType.Text := '';
  SetListFilename('');
  ClearList();

  Silent := SDUCommandLineSwitch('silent');
  VerySilent := SDUCommandLineSwitch('verysilent');
  if VerySilent then
    begin
    Silent := TRUE;
    end;
  self.Visible := not(Silent);
  Application.ShowMainForm := not(Silent);

  if not(Silent) then
    begin
    ProgressBar.Parent := StatusBar;
    end;

  if SDUCommandLineParameter('load', loadFile) then
    begin
    if not(SDUCommandLineParameter('hmac', hmacKey)) then
      begin
      hmacKey := '';
      end;
    if not(SaveLoadList(loadFile, FALSE, hmacKey)) then
      begin
      MessageDlg('The list failed to load correctly.', mtError, [mbOK], 0);
      end
    else
      begin
      UpdateDisplayStatus();
      if SDUCommandLineSwitch('check') then
        begin
        actCheckAllExecute(nil);
        end;  // if SDUCommandLineSwitch('check') then

      end;  // ELSE PART - if not(SaveLoadList(GetListFilename(), FALSE)) then
    end;  // if SDUCommandLineParameter('load', loadFile) then


  if Silent or SDUCommandLineSwitch('quit') then
    begin
    Application.Terminate();
    end;

end;

function TFrmMain.AddFile(filename: string): boolean;
var
  retval: boolean;
  fileSize: int64;
  fileHash: string;
begin
  retval := FALSE;

  fileSize := GetSize(filename);
  if (fileSize > -1) then
    begin
    fileHash := GetChecksumFile(filename);
    if (fileHash <> '') then
      begin
      Filenames.Add(filename);
      FileSizes.Add(inttostr(fileSize));
      FileHashes.Add(fileHash);
      FileStatus.Add(statusOK);

      AddToDisplay(
                   filename,
                   inttostr(fileSize),
                   fileHash,
                   statusOK
                  );

      UpdateDisplayStatus();

      SetFileChanged(TRUE);

      retval := TRUE;
      end;

    end;

  Result := retval;
end;


procedure TfrmMain.SetFileChanged(changed: boolean);
begin
  fFileChanged := changed;

end;

function TfrmMain.GetFileChanged(): boolean;
begin
  Result := fFileChanged;
end;


procedure TfrmMain.AddToDisplay(filename, size, hash, status: string);
var
  addone: boolean;
begin
  if not(Silent) then
    begin
    // This is clunky, but if there are no items in the list, then there will
    // still be a single blank row
    addone := TRUE;
    if (sgFilesList.RowCount = 2) then
      begin
      addone := (sgFilesList.Cells[colFilename , 1] <> '');
      end;

    if addone then
      begin
      sgFilesList.RowCount := sgFilesList.RowCount + 1;
      end;

    sgFilesList.Cells[colFilename, (sgFilesList.RowCount-1)] := filename;
    sgFilesList.Cells[colSize,     (sgFilesList.RowCount-1)] := Format('%.0n', [strtoint64(size) * 1.0]);
    sgFilesList.Cells[colStatus,   (sgFilesList.RowCount-1)] := status;
    end;

end;



// Returns -1 on failure
function TfrmMain.GetSize(filename: string): int64;
var
  retval: int64;
  fileHandle: THandle;
  high, low: DWORD;
begin
  retval := -1;
  
  fileHandle := CreateFile(
                           PChar(filename),  // pointer to name of the file
                           GENERIC_READ,  // access (read-write) mode
                           FILE_SHARE_READ,  // share mode
                           nil,  // pointer to security attributes
                           OPEN_EXISTING, // how to create
                           FILE_ATTRIBUTE_NORMAL,  // file attributes
                           0  // handle to file with attributes to copy
                          );
  if (fileHandle <> INVALID_HANDLE_VALUE) then
    begin
    low := GetFileSize(
                       fileHandle,  // handle of file to get size of
                       @high  // address of high-order word for file size
                      );
    if not((low = $FFFFFFFF) and (GetLastError<>NO_ERROR)) then
      begin
      retVal := high and ($EFFFFFFF);
      retVal := retVal shl 32;
      retVal := retVal or low;
      end;

    CloseHandle(fileHandle);
    end;


  Result := retval;
end;


// Returns '' on failure
function TfrmMain.GetChecksumFile(filename: string): string;
var
  retval: string;
  hashValue: THashArray;
  mac: TMACValue;
begin
  retval := '';

  if (GetListHashType <> '') then
    begin
    if (Key <> '') then
      begin
      // Generate HMAC of file
      mac := TMACValue.Create();
      try
        if HMACFile(Key, filename, hasher, mac) then
          begin
          retval := mac.ValueAsASCIIHex;
          end;
      finally
        mac.Free();
      end;

      end
    else
      begin
      // Generate hash of file
      if Hasher.HashFile(filename, hashValue) then
        begin
        retval := hasher.HashToDisplay(hashValue);
        end;
      end;

    end;

  Result := retval;
end;

function TfrmMain.GetChecksumString(data: string): string;
begin
  Result := GetChecksumString(data, Key);
end;

function TfrmMain.GetChecksumString(data: string; checksumKey: string): string;
var
  retval: string;
  hashValue: THashArray;
  mac: TMACValue;
begin
  retval := '';

  if (GetListHashType <> '') then
    begin
    if (Key <> '') then
      begin
      mac := TMACValue.Create();
      try
        // Generate HMAC of file
        if HMACString(checksumKey, data, hasher, mac) then
          begin
          retval := mac.ValueAsASCIIHex;
          end;
      finally
        mac.Free();
      end;

      end
    else
      begin
      // Generate hash of file
      hashValue := Hasher.HashString(data);
        begin
        retval := hasher.HashToDisplay(hashValue);
        end;
      end;

    end;

  Result := retval;
end;

procedure TfrmMain.ClearList();
var
  x, y: integer;
begin
  Filenames.Clear();
  FileSizes.Clear();
  FileHashes.Clear();
  FileStatus.Clear();

  if not(Silent) then
    begin
    sgFilesList.RowCount := 2;
    sgFilesList.ColCount := 3;
    sgFilesList.FixedRows := 1;

    for x:=0 to (sgFilesList.ColCount-1) do
      begin
      for y:=0 to (sgFilesList.RowCount-1) do
        begin
        sgFilesList.Cells[x, y] := '';
        end;
      end;

    sgFilesList.Cells[colFilename, 0] := 'Filename';
    sgFilesList.Cells[colSize,     0] := 'Size';
    sgFilesList.Cells[colStatus,   0] := 'Status';

    sgFilesList.ColWidths[colFilename] := 512;
    sgFilesList.ColWidths[colSize]     := 96;
    sgFilesList.ColWidths[colStatus]   := 64;

    sgFilesList.DefaultRowHeight := sgFilesList.font.size*2;

    UpdateDisplayStatus();
    end;

  EnableDisable();

end;



// Strip off and return the first line from the TStringList suplied
function TfrmMain.StripLine(var lines: TStringList): string;
var
  retval: string;
begin
  retval:= '';
  if (lines.count>0) then
    begin
    retval:= lines[0];
    lines.Delete(0);
    end;

  Result := retval;
end;


// Get the named value from the <name>=<value> pair
// Note: "name" is case insensitive
// Returns '' if name isn't the name on the line
function TfrmMain.GetValue(name: string; nameValue: string): string;
var
  retval: string;
  tmpName: string;
  value: string;
begin
  retval := '';

  if SDUSplitString(nameValue, tmpName, value, '=') then
    begin
    if (uppercase(name) = uppercase(tmpName)) then
      begin
      retval := value;
      end;

    end;

  Result := retval;
end;


// Save/Load a list
// doSave - Set to TRUE to save the current list; set to FALSE to load
// hmacKey - If set to '', and an HMAC key is required, the user will be
//           prompted for this information
function TfrmMain.SaveLoadList(filename: string; doSave: boolean; hmacKey: string): boolean;
var
  i: integer;
  saveFile: TStringList;
  filecount: integer;
  allOK: boolean;
  tmpVersion: string;
  tmpFilename: string;
  tmpSize: string;
  tmpHash: string;
  strRandom: string;
  tmpHMAC: string;
  calcHMAC: string;
begin
  allOK:= TRUE;

  // Disable the display to operate faster...
  sgFilesList.visible := FALSE;

  if doSave then
    begin
    StartOperation('Saving...');
    end
  else
    begin
    StartOperation('Loading...');
    end;


  saveFile:= TStringList.Create();
  try

    // If loading, read in the file
    if not(doSave) then
      begin
      saveFile.LoadFromFile(filename);
      end;

    // Process basic information...
    if doSave then
      begin
      filecount := Filenames.count;
      // Version hardcoded; the only one that's valid at time of writing
      saveFile.Add('Version=2.00');
      saveFile.Add('Hash='+GetListHashType());
      saveFile.Add('FileCount='+inttostr(filecount));

      // If HMAC, generate random string and HMAC it as a checksum for when the
      // file is opened
      // 1024 - should ideally be the number of bits in the hash?
      strRandom := '';
      calcHMAC := '';
      if (Key <> '') then
        begin
        strRandom := GenerateRandomString(1024);
        calcHMAC := GetChecksumString(strRandom);
        end;
      saveFile.Add('KeyCheckRandom='+strRandom);
      saveFile.Add('KeyCheckHMAC='+calcHMAC);

      saveFile.Add('');
      end
    else
      begin
      // Throw away the file version - there's only one that's valid atm
      tmpVersion := GetValue('Version', StripLine(saveFile));
      if (tmpVersion = '2.00') then
        begin
        // Nothing special... We process any HMAC after getting the hash
        // details
        end
      else if (tmpVersion <> '1.00') then
        begin
        // If it's not v2.00 and not v1.00 - it's unsupported
        allOK := FALSE;
        end;
      allOK:= allOK AND SetListHashType(GetValue('Hash', StripLine(saveFile)));
      filecount := strtoint(GetValue('FileCount', StripLine(saveFile)));

      // If it's a v2.0 checksum file, read in the random data and random data
      // HMAC from the file (used to verify the user's entered the right
      // password)
      Key := '';
      if (tmpVersion = '2.00') then
        begin
        // File is of HMAC supporting version; check HMAC, if needed
        strRandom := GetValue('KeyCheckRandom', StripLine(saveFile));
        tmpHMAC := GetValue('KeyCheckHMAC', StripLine(saveFile));

        if (
            (strRandom <> '') and
            (tmpHMAC <> '')
           ) then
          begin
          Key := hmacKey;
          if (Key = '') then
            begin
            allOK := allOK AND GetKeyFromUser(Key);
            end;
            
          if allOK then
            begin
            // Generate HMAC of random data
            calcHMAC := GetChecksumString(strRandom);

            // Check generated HMAC with file's HMAC
            if (calcHMAC <> tmpHMAC) then
              begin
              MessageDlg('The key you entered does not appear to match that used to create this file.', mtError, [mbOK], 0);
              allOK := FALSE;
              end;

            end;

          end;

        end;

      // Lose blank line
      StripLine(saveFile);
      end;


    if not(doSave) then
      begin
      ClearList();
      end;

    ProgressBarReset(filecount);

    // Process each file in list...
    if (allOK) then
      begin
      for i:=0 to (filecount-1) do
        begin
        if doSave then
          begin
          saveFile.Add('Filename='+Filenames[i]);
          saveFile.Add('FileSize='+FileSizes[i]);
          saveFile.Add('FileHash='+FileHashes[i]);
          saveFile.Add('');
          end
        else
          begin
          tmpFilename:= GetValue('Filename', StripLine(saveFile));
          tmpSize    := GetValue('FileSize', StripLine(saveFile));
          tmpHash    := GetValue('FileHash', StripLine(saveFile));
          StripLine(saveFile);

          if (
              (tmpFilename = '') or
              (tmpSize = '')     or
              (tmpHash = '')
             ) then
            begin
            allOK := FALSE;
            break;
            end
          else
            begin
            Filenames.Add(tmpFilename);
            FileSizes.Add(tmpSize);
            FileHashes.Add(tmpHash);
            FileStatus.Add(statusUnknown);

            AddToDisplay(
                         tmpFilename,
                         tmpSize,
                         tmpHash,
                         statusUnknown
                        );
            end;  // ELSE PART - if (...

          end;  // ELSE PART - if doSave then

        ProgressBarTick();
        end;  // for i:=0 to (filecount-1) do
      end;


    // If saving, write out the file
    if doSave then
      begin
      saveFile.SaveToFile(filename);
      end;

    if (allOK) then
      begin
      SetListFilename(filename)
      end;

  finally
    saveFile.Free();
  end;

  SetFileChanged(FALSE);

  // Reenable display...
  sgFilesList.visible := TRUE;
  SetFocusedControl(sgFilesList);

  EndOperation();

  FLoaded := allOK;

  UpdateDisplayStatus();
  EnableDisable();

  Result := allOK;
end;


function TfrmMain.UserConfirmCancelOp(): boolean;
begin
  CancelFlag := FALSE;
  Result := (MessageDlg('Do you wish to cancel the current operation?', mtConfirmation, [mbYes,mbNo], 0) = mrYes);
end;


procedure TfrmMain.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
var
  dlgReply: DWORD;
  exitFlag: boolean;
begin
  exitFlag := TRUE;

  if RunningOperation() then
    begin
    CancelFlag := TRUE;
    exitFlag := FALSE;
    end
  else if GetFileChanged() then
    begin
    dlgReply := MessageDlg(
               'File list changed since last save - do you wish to save '+CRLF+
               'changes?',
               mtConfirmation,
               [mbYes, mbNo, mbCancel],
               0
              );
    if (dlgReply = mrYes) then
      begin
      exitFlag := Save();
      end
    else if (dlgReply = mrCancel) then
      begin
      exitFlag := FALSE;
      end;

    end;

    
  if exitFlag then
    begin
    Filenames.Free();
    FileSizes.Free();
    FileHashes.Free();
    FileStatus.Free();

    fReportBody.Free();

    Hasher.Free();
    end;

  CanClose := exitFlag;

end;

procedure TfrmMain.miSaveClick(Sender: TObject);
begin
  Save();

end;

procedure TfrmMain.miSaveAsClick(Sender: TObject);
begin
  SaveAs();

end;

procedure TfrmMain.miOpenClick(Sender: TObject);
begin
  Open();

end;


function TfrmMain.Save(): boolean;
var
  retval: boolean;
begin
  if (GetListFilename = '') then
    begin
    retval := SaveAs();
    end
  else
    begin
    SaveLoadList(GetListFilename(), TRUE);
    retval := TRUE;
    end;

  Result:= retval;
end;


function TfrmMain.SaveAs(): boolean;
var
  retval: boolean;
begin
  retval := FALSE;
  SDUOpenSaveDialogSetup(SaveDialog, GetListFilename());
  if SaveDialog.Execute() then
    begin
    SaveLoadList(SaveDialog.Filename, TRUE);
    retval := TRUE;
    end;

  Result:= retval;
end;

procedure TfrmMain.Open();
var
  dlgReply: DWORD;
  openFlag: boolean;
begin
  openFlag := TRUE;

  if GetFileChanged() then
    begin
    dlgReply := MessageDlg(
               'File list changed since last save - do you wish to save '+CRLF+
               'changes?',
               mtConfirmation,
               [mbYes, mbNo, mbCancel],
               0
              );
    if (dlgReply = mrYes) then
      begin
      openFlag := Save();
      end
    else if (dlgReply = mrCancel) then
      begin
      openFlag := FALSE;
      end;

    end;

  if openFlag then
    begin
    SDUOpenSaveDialogSetup(OpenDialog, GetListFilename());
    if OpenDialog.Execute() then
      begin
      if not(SaveLoadList(OpenDialog.Filename, FALSE)) then
        begin
        MessageDlg('The list failed to load correctly.', mtError, [mbOK], 0);
        end;

      UpdateDisplayStatus();
      end;
    end;

end;




procedure TfrmMain.sgFilesListDrawCell(Sender: TObject; ACol,
  ARow: Integer; Rect: TRect; State: TGridDrawState);
const
  LM = 3; {each indiviual cell's left margin}
  TM = 2; {each indiviual cell's top margin}
var
  sgrObj: TStringGrid;
  canvas: TCanvas;
  bgColor: TColor;
  align: TAlignment;
begin
  sgrObj := TStringGrid(Sender);
  canvas := sgrObj.Canvas;

  // Different colours and alignments...
  align := taLeftJustify;
  if (ARow > 0) and (ACol = colSize) then
    begin
    align := taRightJustify;
    end;

  if (ARow > 0) and (ACol = colStatus) then
    begin
    bgColor := canvas.Brush.Color;

    if (sgrObj.Cells[ACol, ARow] = statusOK) then
      begin
      bgColor := clLime;
      end
    else if (sgrObj.Cells[ACol, ARow] = statusChanged) then
      begin
      bgColor := clRed;
      end
    else if (sgrObj.Cells[ACol, ARow] = statusDeleted) then
      begin
      bgColor := clRed;
      end
    else if (sgrObj.Cells[ACol, ARow] = statusUnchecked) then
      begin
      bgColor := clBlue;
      end;

    canvas.Brush.Color := bgColor;
    end;

  canvas.FillRect(Rect);  { clear the rectangle }
  //  canvas.TextRect(Rect, Rect.Left + LM, Rect.Top + TM, sgrObj.Cells[ACol, ARow]);
  DrawTheText(sgrObj, Rect, sgrObj.Cells[ACol, ARow], align);
  
end;


// Note: This is a modified version of the function potsed 1998/07/06 by "s...@my-dejanews.com", found on Google
procedure TfrmMain.DrawTheText(AGrid: TStringGrid; ARect: TRect; AValue: string; Alignment: TAlignment);
var
  canvas: TCanvas;
  canvasHandle: THandle;
  horzOffset: integer;
  vertOffset: integer;
begin
  canvas := AGrid.Canvas;
  canvasHandle := canvas.Handle;

  vertOffset := (((ARect.Bottom - ARect.Top) - canvas.TextExtent(AValue).CY) div 2);
  horzOffset := canvas.TextExtent('Mi').CX div 4;
  case Alignment of
    taLeftJustify:
      begin
      SetTextAlign(canvasHandle, TA_LEFT or TA_TOP or TA_NOUPDATECP);
      ExtTextOut(
                 canvasHandle,
                 ARect.Left + horzOffset,
                 ARect.Top + vertOffset,
                 ETO_CLIPPED or ETO_OPAQUE,
                 @ARect,
                 PChar(AValue),
                 Length(AValue),
                 nil
                );
      end;

    taRightJustify:
      begin
      SetTextAlign(canvasHandle, TA_RIGHT or TA_TOP or TA_NOUPDATECP);
      ExtTextOut(
                 canvasHandle,
                 ARect.Right - horzOffset,
                 ARect.Top + vertOffset,
                 ETO_CLIPPED or ETO_OPAQUE,
                 @ARect,
                 PChar(AValue),
                 Length(AValue),
                 nil
                );
      end;
        
    taCenter:
      begin
      horzOffset := ((ARect.Right - ARect.Left) - canvas.TextExtent (AValue).CX) div 2;
      SetTextAlign(canvasHandle, TA_LEFT or TA_TOP or TA_NOUPDATECP);
      ExtTextOut(
                 canvasHandle,
                 ARect.Left + horzOffset,
                 ARect.Top + vertOffset,
                 ETO_CLIPPED or ETO_OPAQUE,
                 @ARect,
                 PChar(AValue),
                 Length(AValue),
                 nil
                );
      end;
    end;

end;


function TfrmMain.GetListFilename(): string;
begin
  Result := fListFilename;
end;

procedure TfrmMain.SetListFilename(filename: string);
begin
  fListFilename := filename;
end;


function TfrmMain.GetListHashType(): string;
begin
  Result := fListHashType;
end;

function TfrmMain.SetListHashType(hashType: string): boolean;
var
  i: fhHashType;
  found: boolean;
begin
  found := FALSE;

  fListHashType := '';

  for i:=low(fhHashType) to high(fhHashType) do
    begin
    Hasher.ActiveHash := i;
    if (Hasher.Title = hashType) then
      begin
      found := TRUE;
      fListHashType := hashType;
      break;
      end;
    end;

    
  if (not(found) and (hashType<>'')) then
    begin
    Hasher.ActiveHash := hashSHA1;
    MessageDlg('Unknown hash algorithm "'+hashType+'"', mtError, [mbOK], 0);
    end;

  Result := found;
end;



procedure TfrmMain.UpdateDisplayStatus();
var
  i: integer;
  cntOK: integer;
  cntChanged: integer;
  cntDeleted: integer;
  cntUnchecked: integer;
  cntUnknown: integer;

  totalBytes: int64;
begin
  if not(Silent) then
    begin
    cntOK:= 0;
    cntChanged:= 0;
    cntDeleted:= 0;
    cntUnchecked:= 0;
    cntUnknown:= 0;

    totalBytes:= 0;

    for i:=0 to (Filenames.Count-1) do
      begin
      if (FileStatus[i] = statusOK) then
        begin
        inc(cntOK);
        end
      else if (FileStatus[i] = statusChanged) then
        begin
        inc(cntChanged);
        end
      else if (FileStatus[i] = statusDeleted) then
        begin
        inc(cntDeleted);
        end
      else if (FileStatus[i] = statusUnchecked) then
        begin
        inc(cntUnchecked);
        end
      else
        begin
        inc(cntUnknown);
        end;

      totalBytes := totalBytes + strtoint(FileSizes[i]);
      end;

    edStatusOK.text        := Format('%.0n', [cntOK * 1.0]);;
    edStatusChanged.text   := Format('%.0n', [cntChanged * 1.0]);;
    edStatusDeleted.text   := Format('%.0n', [cntDeleted * 1.0]);;
    edStatusUnchecked.text := Format('%.0n', [cntUnchecked * 1.0]);;
    edStatusUnknown.text   := Format('%.0n', [cntUnknown * 1.0]);;

    edStatusTotal.text     := Format('%.0n', [Filenames.Count * 1.0]);;

    edBytesTotal.text := Format('%.0n', [totalBytes * 1.0]);;

    edHashType.Text := '';
    if (FLoaded) then
      begin
      if (Key = '') then
        begin
        edHashType.Text := Hasher.Title;
        end
      else
        begin
        edHashType.Text := 'HMAC: '+Hasher.Title;
        end;
      end;

    self.Caption := Application.Title;
    if (GetListFilename() <> '') then
      begin
      self.Caption := self.Caption + ' - ' + GetListFilename();
      end;

    if GetFileChanged() then
      begin
      self.Caption := self.Caption + ' *';
      end;

    EnableDisable();
    Application.ProcessMessages();
    end;

end;


procedure TfrmMain.StartOperation(operation: string);
begin
  CancelFlag := FALSE;
  
  if not(Silent) then
    begin
    Screen.Cursor := crAppStart;

    if (operation = '') then
      begin
      operation := operationReady;
      end;

    statusbar.Panels[0].Text := operation;
    EnableDisable();
    end;

  fReportLastOpStartTime:= Now();
end;


// Returns TRUE if an operation is currently running, otherwise FALSE
function TfrmMain.RunningOperation(): boolean;
begin
  Result := (statusbar.Panels[0].Text <> operationReady);
end;


procedure TfrmMain.EndOperation();
begin
  fReportLastOpEndTime:= Now();

  if not(Silent) then
    begin
    Screen.Cursor := crArrow;
    statusbar.Panels[0].Text := operationReady;
    ProgressBar.Max := 100;
    ProgressBar.Position := 0;
    EnableDisable();
    end;

end;


procedure TfrmMain.DeleteRow(row: integer);
var
  x, y: integer;
begin
  for y:=row to sgFilesList.RowCount - 2 do
    begin
    for x:=0 to sgFilesList.ColCount - 1 do
      begin
      sgFilesList.Cells[x, y] := sgFilesList.Cells[x, (y+1)];
      end;
    end;

  sgFilesList.RowCount := sgFilesList.RowCount - 1;
end;


procedure TfrmMain.EnableDisable();
var
  i: integer;
  disableAll: boolean;
begin
  if not(Silent) then
    begin
    disableAll := RunningOperation() OR (GetListHashType() = '');

    // If carrying out an operation, prevent user from doing anything
    for i:=0 to (MainMenu.Items.count-1) do
      begin
      MainMenu.Items[i].enabled := not(disableAll);
      end;


    miNew.Enabled := not(RunningOperation());
    miOpen.Enabled := not(RunningOperation());
    miExit.Enabled := not(RunningOperation());
    miFile.Enabled := not(RunningOperation());
    miSave.Enabled := not(RunningOperation()) AND (GetListHashType() <> '');
    miSaveAs.Enabled := not(RunningOperation()) AND (GetListHashType() <> '');
    miView.Enabled := not(RunningOperation());
    miOptions.Enabled := not(RunningOperation());
    miHelp.Enabled := not(RunningOperation());

    
    // Specific items...
    actRemoveSelected.enabled := ( (Filenames.Count >= 1) and (sgFilesList.Selection.Top>=1) );

    actCheckSelected.enabled := ( (Filenames.Count >= 1) and (sgFilesList.Selection.Top>=1) );
    actCheckAll.enabled := (Filenames.Count >= 1);

    miRecalcSelected.enabled := ( (Filenames.Count >= 1) and (sgFilesList.Selection.Top>=1) );
    miRecalcAll.enabled := (Filenames.Count >= 1);

    actProperties.enabled := (
                              (Filenames.Count >= 1) and
                              (sgFilesList.Selection.Top >= 1) and
                              (sgFilesList.Selection.Top = sgFilesList.Selection.Bottom)
                             );

    miReport.enabled := (fReportTitle<>'') and not(disableAll);
    end;

end;

procedure TfrmMain.RecalcSelectedClick(Sender: TObject);
begin
  RecalcRange(sgFilesList.Selection.Top-1, sgFilesList.Selection.Bottom-1);

end;

procedure TfrmMain.RecalcAllClick(Sender: TObject);
begin
  RecalcRange(0, Filenames.Count-1);

end;


// startIdx - 0-based index; 0 being the first filename
// endIdx - 0-based index; (Filenames.count-1) being the last possible filename
procedure TfrmMain.CheckRange(startIdx, endIdx: integer);
var
  i: integer;
  int64FileSize: int64;
  strFileSize: string;
  fileHash: string;
  currStatus: string;
  filesChanged: TStringList;
  filesDeleted: TStringList;
  filesUnchecked: TStringList;
  checkedOKCnt: integer;
begin
  checkedOKCnt:= 0;

  // Set the file's so their status is "unknown"
  for i:=startIdx to endIdx do
    begin
    FileStatus[i] := statusUnknown;
    // Refresh display, if not running quiet
    if not(Silent) then
      begin
      sgFilesList.Cells[colStatus, i+1] := statusUnknown;
      end;

    end;


  filesChanged:= TStringList.Create();
  try
    filesDeleted:= TStringList.Create();
    try
      filesUnchecked:= TStringList.Create();
      try

      ProgressBarReset(endIdx - startIdx + 1);

      StartOperation('Checking...');

      // Process the files
      for i:=startIdx to endIdx do
        begin
        currStatus := statusUnknown;

        if not(FileExists(Filenames[i])) then
          begin
          filesDeleted.Add(Filenames[i]);
          currStatus := statusDeleted;
          end
        else
          begin
          int64FileSize := GetSize(Filenames[i]);
          if (int64FileSize < 0) then
            begin
            filesUnchecked.Add(Filenames[i]);
            currStatus := statusUnchecked;
            end
          else
            begin
            strFileSize := inttostr(int64FileSize);
            if (strFileSize <> FileSizes[i]) then
              begin
              filesChanged.Add(Filenames[i]);
              currStatus := statusChanged;
              end
            else
              begin
              fileHash := GetChecksumFile(Filenames[i]);
              if (fileHash = '') then
                begin
                currStatus := statusUnchecked;
                end
              else
                begin
                if (fileHash <> FileHashes[i]) then
                  begin
                  filesChanged.Add(Filenames[i]);
                  currStatus := statusChanged;
                  end
                else
                  begin
                  inc(checkedOKCnt);
                  currStatus := statusOK;
                  end;  // ELSE PART - if (fileHash <> FileHashes[i])
                end;  // ELSE PART - if (fileHash = '')
              end;  // ELSE PART - if (strFileSize <> FileSizes[i])
            end;  // ELSE PART - if (int64FileSize < 0)
          end;  // ELSE PART - if not(FileExists(Filenames[i]))


        FileStatus[i] := currStatus;

        // Refresh display, if not running quiet
        if not(Silent) then
          begin
          sgFilesList.Cells[colStatus, i+1] := currStatus;
          UpdateDisplayStatus();
          end;

        ProgressBarTick();
        if CancelFlag then
          begin
          if UserConfirmCancelOp() then
            begin
            break;
            end;
          end;

        end;  // for i:=startIdx to endIdx do


        EndOperation();

        ReportNew('Check Integrity');

        ReportAddFileList(
                       filesUnchecked,
                       'Unchecked files',
                       'could not be checked'
                      );

        ReportAddFileList(
                       filesDeleted,
                       'Deleted files',
                       'appear to have been deleted'
                      );

        ReportAddFileList(
                       filesChanged,
                       'Changed files',
                       'have changed'
                      );

        // If no problems...
        if (
            (filesUnchecked.count=0) and
            (filesDeleted.count=0) and
            (filesChanged.count=0)
           ) then
          begin
          // If running "very silent" - we don't display a dialog unless there
          // were problems
          if not(VerySilent) then
            begin
            ReportAddLine(inttostr(checkedOKCnt)+' file checked succesfully');
            ReportAddLine('');
            ReportAddTimeTaken();
            ReportShow();
            end;
          end
        else
          begin
          // One or more problems...
          ReportAddTimeTaken();
          ReportShow();
          end;

        EnableDisable(); // Required for the "View|Report..." menuitem

      finally
        filesUnchecked.Free();
      end;
    finally
      filesDeleted.Free();
    end;
  finally
    filesChanged.Free();
  end;

end;


procedure TfrmMain.ReportNew(title: string);
begin
  fReportTitle := title;
  fReportBody.Clear();
end;

procedure TfrmMain.ReportAddFileList(files: TStringList; title: string; msgSeg: string);
begin
  if (files.count>0) then
    begin
    fReportBody.Add('The following '+inttostr(files.count)+' files '+msgSeg+':');
    fReportBody.AddStrings(files);
    fReportBody.Add('');
    end;

end;

procedure TfrmMain.ReportAddLine(line: string);
begin
  fReportBody.Add(line);

end;

procedure TfrmMain.ReportAddTimeTaken();
var
  timeDiff: TDateTime;
begin
  timeDiff := fReportLastOpEndTime - fReportLastOpStartTime;
  fReportBody.Add('Time taken: '+TimeToStr(timeDiff));

end;

procedure TfrmMain.ReportShow();
var
  frmReport: TfrmReport;
begin
  if (fReportBody.Count = 0) then
    begin
    ReportAddLine('The operation completed successfully.');
    ReportAddLine('');
    end;

  frmReport:= TfrmReport.Create(nil);
  try
    frmReport.SetTitle(fReportTitle);
    frmReport.SetReport(fReportBody);
    frmReport.ShowModal();
  finally
    frmReport.Free();
  end;

end;


// startIdx - 0-based index; 0 being the first filename
// endIdx - 0-based index; (Filenames.count-1) being the last possible filename
procedure TfrmMain.RecalcRange(startIdx, endIdx: integer);
var
  i: integer;
  int64FileSize: int64;
  fileHash: string;
  currStatus: string;
  filesDeleted: TStringList;
  filesUnprocessed: TStringList;
  recalcOKCnt: integer;
begin
  recalcOKCnt:= 0;

  // Set the file's so their status is "unknown"
  for i:=startIdx to endIdx do
    begin
    FileStatus[i] := statusUnknown;
    // Refresh display, if not running quiet
    if not(Silent) then
      begin
      sgFilesList.Cells[colStatus, i+1] := statusUnknown;
      end;

    end;


  filesDeleted:= TStringList.Create();
  try
    filesUnprocessed:= TStringList.Create();
    try

    ProgressBarReset(endIdx - startIdx + 1);

    StartOperation('Recalculating...');

    // Process the files
    for i:=startIdx to endIdx do
      begin
      currStatus := statusUnknown;

      if not(FileExists(Filenames[i])) then
        begin
        filesDeleted.Add(Filenames[i]);
        currStatus := statusDeleted;
        end
      else
        begin
        int64FileSize := GetSize(Filenames[i]);
        if (int64FileSize < 0) then
          begin
          filesUnprocessed.Add(Filenames[i]);
          currStatus := statusUnchecked;
          end
        else
          begin
          fileHash := GetChecksumFile(Filenames[i]);
          if (fileHash = '') then
            begin
            filesUnprocessed.Add(Filenames[i]);
            currStatus := statusUnchecked;
            end
          else
            begin
            FileSizes[i] := inttostr(int64FileSize);
            FileHashes[i] := fileHash;
            inc(recalcOKCnt);
            currStatus := statusOK;
            end;  // ELSE PART - if (fileHash = '')
          end;  // ELSE PART - if (int64FileSize < 0)
        end;  // ELSE PART - if not(FileExists(Filenames[i]))


      FileStatus[i] := currStatus;

      // Refresh display, if not running quiet
      if not(Silent) then
        begin
        sgFilesList.Cells[colStatus, i+1] := currStatus;
        UpdateDisplayStatus();
        end;

      ProgressBarTick();
      if CancelFlag then
        begin
        if UserConfirmCancelOp() then
          begin
          break;
          end;
        end;

      end;  // for i:=startIdx to endIdx do


    EndOperation();

    ReportNew('Recalculate Integrity');

    ReportAddFileList(
                   filesUnprocessed,
                   'Unprocessed files',
                   'could not be processed'
                  );

    ReportAddFileList(
                   filesDeleted,
                   'Deleted files',
                   'appear to have been deleted'
                  );

      if (
          (filesUnprocessed.count=0) and
          (filesDeleted.count=0)
         ) then
        begin
        ReportAddLine(inttostr(recalcOKCnt)+' file(s) recalculated succesfully');
        ReportAddLine('');
        end;

      ReportAddTimeTaken();
      ReportShow();
      EnableDisable(); // Required for the "View|Report..." menuitem

    finally
      filesUnprocessed.Free();
    end;
  finally
    filesDeleted.Free();
  end;

end;


procedure TfrmMain.miAboutClick(Sender: TObject);
var
  aboutDlg: TAbout_F;
begin
  aboutDlg := TAbout_F.Create(nil);
  try
    aboutDlg.ShowModal();
  finally
    aboutDlg.Free();
  end;

end;

procedure TfrmMain.FormShow(Sender: TObject);
begin
  if not(Silent) then
    begin
    pnlLine1.Caption := '';
    pnlLine2.Caption := '';
    pnlLine3.Caption := '';

    // 25 should be large enough...
    self.Constraints.MinWidth := gbStatus.Width + 25;
    self.Constraints.MinHeight := 300;

    EndOperation();
    end;

end;

procedure TfrmMain.sgFilesListClick(Sender: TObject);
begin
  CancelFlag := TRUE;
  EnableDisable();
  
end;


procedure TfrmMain.ProgressBarReset(max: int64);
begin
  if not(Silent) then
    begin
    ProgressBar.Max := max;
    ProgressBar.Position := 0;
    Application.ProcessMessages();
    end;

end;

procedure TfrmMain.ProgressBarTick();
begin
  if not(Silent) then
    begin
    ProgressBar.Position := ProgressBar.Position + 1;
    Application.ProcessMessages();
    end;

end;


procedure TfrmMain.statusBarDrawPanel(StatusBar: TStatusBar;
  Panel: TStatusPanel; const Rect: TRect);
begin
  if Panel = StatusBar.Panels[1] then
    begin
    ProgressBar.Left := rect.Left;
    ProgressBar.Top := rect.Top;
    // We drop the width by pbTransferred.height to allow for the triangle in
    // the corner
    // ONLY DO THIS IF THE PROGRESS BAR'S STATUS PANEL IS THE ONE ON THE FAR
    // RIGHT
    ProgressBar.Width := rect.Right - ProgressBar.Left - ProgressBar.height;
    ProgressBar.Height := rect.Bottom - ProgressBar.Top;
    end;

end;


procedure TfrmMain.sgFilesListDblClick(Sender: TObject);
begin
  if ( (Filenames.Count >= 1) and (sgFilesList.Selection.Top>=1) and (sgFilesList.Selection.Top = sgFilesList.Selection.Bottom) ) then
    begin
    actPropertiesExecute(nil);
    end;

end;


procedure TfrmMain.miNewClick(Sender: TObject);
var
  dlgReply: DWORD;
  newFlag: boolean;
  selectDlg: TfrmSelectHash;
  selectedHash: string;
  enteredKey: string;
begin
  newFlag := TRUE;

  if GetFileChanged() then
    begin
    dlgReply := MessageDlg(
               'File list changed since last save - do you wish to save '+CRLF+
               'changes?',
               mtConfirmation,
               [mbYes, mbNo, mbCancel],
               0
              );
    if (dlgReply = mrYes) then
      begin
      newFlag := Save();
      end
    else if (dlgReply = mrCancel) then
      begin
      newFlag := FALSE;
      end;

    end;


  if newFlag then
    begin
    selectDlg:= TfrmSelectHash.Create(nil);
    try
      newFlag := (selectDlg.ShowModal() = mrOK);
      selectedHash := selectDlg.SelectedHash;
      enteredKey := selectDlg.Key;
    finally
      selectDlg.Free();
    end;


    if newFlag then
      begin
      FLoaded := TRUE;
      Key := enteredKey;
      SetListHashType(selectedHash);
      SetListFilename('');
      ClearList();

      UpdateDisplayStatus();
      end;

    end;

end;


procedure TfrmMain.miReportClick(Sender: TObject);
begin
  ReportShow();

end;

procedure TfrmMain.FormResize(Sender: TObject);
begin
  if not(Silent) then
    begin
    gbStatus.Left := (self.width - gbStatus.width) div 2;
    end;

end;

procedure TfrmMain.miOptionsClick(Sender: TObject);
var
  dlg: TfrmOptions;
begin
  dlg:= TfrmOptions.Create(nil);
  try
    dlg.ShowModal();
  finally
    dlg.Free();
  end;

end;

procedure TfrmMain.FormClick(Sender: TObject);
begin
  CancelFlag := TRUE;

end;

// Generate a random string with at least the specified number of bits entropy
function TfrmMain.GenerateRandomString(minBits: integer): string;
const
  CANDIDATE_CHARS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890!$%^&*()_-+=[]{};#:@~,./<>?\|';
var
  i: integer;
  idx: integer;
  bitsPerChar: double;
  minCharLen: integer;
  retVal: string;
begin
  // The RNG used to generate this string is *not* critical; we store it in
  // plaintext in the file.
  // This *isn't* a security risk, as it doens't need to be protected

  bitsPerChar:= (log2(length(CANDIDATE_CHARS)));
  minCharLen := trunc(minBits / bitsPerChar) + 1;

  retVal := '';
  for i:=1 to minCharLen do
    begin
    // +1 because we index from 1
    idx := random(length(CANDIDATE_CHARS)) + 1;
    retVal := retVal + CANDIDATE_CHARS[idx];
    end;
    
  Result := retVal;
end;


function TfrmMain.GetKeyFromUser(var userKey: string): boolean;
var
  frm: TfrmKeyEntry;
  retVal: boolean;
begin
  retVal := FALSE;

  frm:= TfrmKeyEntry.Create(nil);
  try
    if (frm.ShowModal = mrOK) then
      begin
      userKey := frm.Key();
      retVal := TRUE;
      end;
  finally
    frm.Free();
  end;

  Result := retVal;
end;



procedure TfrmMain.actAddFileExecute(Sender: TObject);
var
  filename: string;
begin
  if OpenDialogAdd.Execute() then
    begin
    filename := OpenDialogAdd.Filename;
    StartOperation('Adding files...');

    if not(AddFile(filename)) then
      begin
      MessageDlg(
                 'Unable to add file:'+CRLF+
                 CRLF+
                 filename+CRLF+
                 CRLF+
                 'Please check and make sure that this file is not in use by '+CRLF+
                 'another application.',
                 mtError,
                 [mbOK],
                 0
                );
      end;

    EndOperation();
    end;

  EnableDisable();
end;



procedure TfrmMain.actAddDirExecute(Sender: TObject);
var
  addDir: string;
  iterator: TSDUFileIterator;
  tmpFilename: string;
  frmAddDir: TfrmAddDir;
  failedToAdd: TStringList;
  totalFiles: integer;
begin
  addDir := '';

  frmAddDir := TfrmAddDir.Create(nil);
  try
    if (frmAddDir.ShowModal = mrOK) then
      begin
      // Disable the display to operate faster...
      sgFilesList.visible := FALSE;

      failedToAdd := TStringList.Create();
      try
        StartOperation('Adding files...');

        iterator:= TSDUFileIterator.Create(nil);
        try
          iterator.Directory := frmAddDir.edDir.text;
          iterator.FileMask := frmAddDir.edFileMask.text;
          iterator.Sorted := TRUE;
          iterator.RecurseSubDirs:= frmAddDir.ckRecursive.checked;
          iterator.OmitStartDirPrefix := FALSE;

          iterator.IncludeDirNames := FALSE;

          iterator.Reset();
          totalFiles := iterator.Count();
          ProgressBarReset(totalFiles);

          iterator.Reset();
          tmpFilename := iterator.Next();
          while (tmpFilename <> '') do
            begin
            if not(AddFile(tmpFilename)) then
              begin
              failedToAdd.Add(tmpFilename);
              end;

            ProgressBarTick();
            if CancelFlag then
              begin
              if UserConfirmCancelOp() then
                begin
                break;
                end;
              end;

            tmpFilename := iterator.Next();
            end;

        finally
          iterator.Free();
        end;

        EndOperation();

        ReportNew('Add directory');
        ReportAddFileList(failedToAdd, 'Files not added', 'could not be added');
        if (failedToAdd.count=0) then
          begin
          ReportAddLine(inttostr(totalFiles)+' file(s) added successfully.');
          ReportAddLine('');
          end;
        ReportAddTimeTaken();
        ReportShow();
        EnableDisable(); // Required for the "View|Report..." menuitem

      finally
        failedToAdd.Free();
      end;

      end;

  finally
    frmAddDir.Free();
  end;                                      

  // Reenable display...
  sgFilesList.visible := TRUE;
  SetFocusedControl(sgFilesList);

end;


procedure TfrmMain.actRemoveSelectedExecute(Sender: TObject);
var
  y: integer;
begin
  if (Filenames.Count >= 1) then
    begin
    if (MessageDlg('Are you sure you wish to to remove the selected files?', mtConfirmation, [mbYes,mbNo], 0) = mrYes) then
      begin
      for y:=sgFilesList.Selection.Bottom downto sgFilesList.Selection.Top do
        begin
        DeleteRow(y);
        Filenames.Delete(y-1);
        FileSizes.Delete(y-1);
        FileHashes.Delete(y-1);
        FileStatus.Delete(y-1);
        end;

      if (sgFilesList.RowCount<2) then
        begin
        ClearList();
        end;

      end;

    end;
    
end;


procedure TfrmMain.actCheckSelectedExecute(Sender: TObject);
begin
  CheckRange(sgFilesList.Selection.Top-1, sgFilesList.Selection.Bottom-1);

end;

procedure TfrmMain.actCheckAllExecute(Sender: TObject);
begin
  CheckRange(0, Filenames.Count-1);

end;

procedure TfrmMain.actPropertiesExecute(Sender: TObject);
var
  propDlg: TfrmProperties;
  y: integer;
begin
  propDlg:= TfrmProperties.Create(nil);
  try
    y := sgFilesList.Selection.Top;

    propDlg.edFilename.text:= Filenames[y-1];
    propDlg.edSize.text:=  Format('%.0n', [strtoint64(FileSizes[y-1]) * 1.0])+' bytes';
    propDlg.edHash.text:= FileHashes[y-1];
    propDlg.edStatus.text:= FileStatus[y-1];

    if (propDlg.edStatus.text = statusOK) then
      begin
      propDlg.edStatus.Color := clLime;
      end
    else if (propDlg.edStatus.text = statusChanged) then
      begin
      propDlg.edStatus.Color := clRed;
      end
    else if (propDlg.edStatus.text = statusUnchecked) then
      begin
      propDlg.edStatus.Color := clBlue;
      end;

    propDlg.ShowModal();
  finally
    propDlg.Free();
  end;

end;

END.



