unit SelectHash;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls,
  HashAlgUnified_U, ComCtrls, PasswordRichEdit;

type
  TfrmSelectHash = class(TForm)
    pbOK: TButton;
    cbHashList: TComboBox;
    pbCancel: TButton;
    ckHMAC: TCheckBox;
    Label2: TLabel;
    Label1: TLabel;
    edKey: TEdit;
    procedure FormShow(Sender: TObject);
    procedure pbOKClick(Sender: TObject);
    procedure ckHMACClick(Sender: TObject);
  private
    fSelectedHash: string;

    function GetSelectedHash(): string;
    function GetKey(): string;

    function GetDispName(hashType: fhHashType): string;
    function GetName(hashType: fhHashType): string;
  public
    property SelectedHash: string read GetSelectedHash;
    property Key: string read GetKey;
  end;


implementation

{$R *.DFM}

uses
  SDUGeneral;

const
  DEFAULT_HASH = 'SHA-1';

procedure TfrmSelectHash.FormShow(Sender: TObject);
var
  i: integer;
  curr: fhHashType;
begin
  for curr:=low(fhHashType) to high(fhHashType) do
    begin
    cbHashList.Items.Add(GetDispName(curr));
    end;

  // Ensure the list is sorted...
  cbHashList.Sorted := FALSE;
  cbHashList.Sorted := TRUE;


  cbHashList.ItemIndex := 0;
  for i:=0 to (cbHashList.Items.count-1) do
    begin
    if (Pos(DEFAULT_HASH, cbHashList.Items[i]) > 0) then
      begin
      cbHashList.ItemIndex := i;
      end;
    end;


  edKey.PasswordChar := '*';
  edKey.Text := '';

end;


function TfrmSelectHash.GetDispName(hashType: fhHashType): string;
var
  hash: THashAlgUnified;
  retval: string;
begin
  hash:= THashAlgUnified.Create(nil);
  try
    hash.ActiveHash := hashType;
    retval := hash.Title;
    retval := retval + ' ('+inttostr(hash.HashLength)+')';
  finally
    hash.Free();
  end;

  Result := retval;
end;


function TfrmSelectHash.GetName(hashType: fhHashType): string;
var
  hash: THashAlgUnified;
  retval: string;
begin
  hash:= THashAlgUnified.Create(nil);
  try
    hash.ActiveHash := hashType;
    retval := hash.Title;
  finally
    hash.Free();
  end;

  Result := retval;
end;


function TfrmSelectHash.GetSelectedHash(): string;
begin
  Result := fSelectedHash;
end;

function TfrmSelectHash.GetKey(): string;
var
  retVal: string;
begin
  if (ckHMAC.checked) then
    begin
    retVal := edKey.text;
    end
  else
    begin
    retVal := '';
    end;

  Result := retVal;
end;


procedure TfrmSelectHash.pbOKClick(Sender: TObject);
var
  i: fhHashType;
  allOK: boolean;
begin
  allOK:= TRUE;

  fSelectedHash := '';

  for i:=low(fhHashType) to high(fhHashType) do
    begin
    if (cbHashList.Items[cbHashList.ItemIndex] = GetDispName(i)) then
      begin
      fSelectedHash := GetName(i);
      end;
    end;

  if (ckHMAC.checked) then
    begin
    if (GetKey() = '') then
      begin
      MessageDlg('You must specify a password to use HMAC checksums', mtError, [mbOK], 0);
      allOK := FALSE;
      end;
    end;

  if allOK then
    begin
    ModalResult := mrOK;
    end;

end;

procedure TfrmSelectHash.ckHMACClick(Sender: TObject);
begin
  SDUEnableControl(edKey, ckHMAC.Checked);
end;

END.


